#!/usr/bin/env python
#
# https://www.pyimagesearch.com/2015/05/25/basic-motion-detection-and-tracking-with-python-and-opencv/
#

import io
import sys
import time
import dill as pickle
import logging
import datetime
from io import BytesIO
from copy import deepcopy
from threading import Thread, Lock
from argparse import ArgumentParser

import cv2
import lz4.frame
import redis
import numpy as np

logging.basicConfig(
    format='%(asctime)s;%(levelname)s;%(message)s',
    #level=logging.DEBUG,
    #level=logging.INFO,
    #level=logging.WARN,
    level=logging.ERROR,
    filename='log.txt'
)

class NoFrameException(Exception): pass



def image_resize(image, width=None, height=None, inter=cv2.INTER_AREA):
    dim = None
    (h, w) = image.shape[:2]
    if width is None and height is None:
        return image
    if width is None:
        r = height / float(h)
        dim = (int(w * r), height)
    else:
        r = width / float(w)
        dim = (width, int(h * r))
    resized = cv2.resize(image, dim, interpolation=inter)
    return resized


class Messager(object):
    def __init__(self):
        self.message = None
        self.readlock = Lock()
        self.writelock = Lock()

    def set(self, message):
        self.writelock.acquire()
        self.message = message
        self.writelock.release()

    def get(self):
        #self.readlock.acquire()
        while self.message is None: time.sleep(0.0001)
        message = self.message # deepcopy(self.message)
        #self.readlock.release()
        return message


class CamReader(object):
    def __init__(self, args, conn):
        self.vs = cv2.VideoCapture(args.camera)
        self.vs.set(cv2.CAP_PROP_FRAME_WIDTH, 1280)
        self.vs.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)    
        time.sleep(3.0)
        self.firstFrame = None
        self.conn = conn

    def __call__(self):
        while True:
            #start = time.time()
            res = self.vs.read()
            #start2 = time.time()
            self.conn.set(res)

    def __del__(self):
        self.vs.release()


class RedisWriter(object):
    def __init__(self, args, conn):
        self.red = redis.Redis(unix_socket_path=args.redissock)
        self.conn = conn
        
    def __call__(self):
        while True:
            #start = time.time()
            key, val = self.conn.get() ####
            #start2 = time.time()
            self.red.set(key, val)

    def publish(self, channel, message):
        self.red.publish(channel, message)

'''
class ProcessImage(object):
    def __init__(self, args, reader, writer):
        self.args = args
        self.reader = reader
        self.writer = writer

    def __call__(self):
'''

            
if __name__ == '__main__':
    parser = ArgumentParser(description='first program for motion detection')
    parser.add_argument('--camera', '-c', type=int, default=0, help='number of camera, eg. 0 for /dev/video0')
    parser.add_argument('--minarea', '-m', type=int, default=500, help='minimum detection area size')
    parser.add_argument('--width', '-w', type=int, default=640, help='image width')
    parser.add_argument('--blur', '-b', type=int, default=21, help='gaussian blur (remove noise)')
    #parser.add_argument('--redishost', '-H', type=str, default='tv', help='the redis-server')
    #parser.add_argument('--redishost', '-H', type=str, default=None, help='the redis-server')
    parser.add_argument('--redissock', '-H', type=str, default='/var/run/redis/redis-server.sock', help='the redis-server')
    args = parser.parse_args()

    cam2proc = Messager()
    proc2red = Messager()
    
    cam = CamReader(args, cam2proc)
    #prc = ProcessImage(args, cam2proc, proc2red)
    red = RedisWriter(args, proc2red)
    print('red =', red)
    
    pc = Thread(target=cam) #, daemon=True)
    #pi = Thread(target=prc) #, daemon=True)
    pr = Thread(target=red) #, daemon=True)
    
    pc.start()
    #pi.start()
    pr.start()


    avg = None
    while True:
        start = time.time()
        
        #ret, frame = vs.read()
        ret, frame = cam2proc.get() ####
        #logging.debug('ProcessImage.reader.recv (time);%0.5f'%(time.time() - start))
        
        detected = False
        orig = frame.copy()
        
        if frame is None:
            raise NoFrameException('Could not read frame, sorry.')

        small = image_resize(frame, width=args.width)
        #small = frame.copy()
            
        gray = cv2.cvtColor(small, cv2.COLOR_BGR2GRAY) # convert colorspace
        gray = cv2.GaussianBlur(gray, (args.blur, args.blur), 0) # remove noise

        if avg is None:
            avg = gray.copy().astype("float")
            continue
        #cv2.accumulateWeighted(gray, avg, 0.5)
        cv2.accumulateWeighted(gray, avg, 0.1)
        frameDelta = cv2.absdiff(gray, cv2.convertScaleAbs(avg))

        thresh = cv2.threshold(frameDelta, 25, 255, cv2.THRESH_BINARY)[1]  # convert grayscale to hard b/w
        thresh = cv2.dilate(thresh, None, iterations=2) # enance threshold
        cnts = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE) # find contours
        if len(cnts) == 2: cnts = cnts[0]
        elif len(cnts) == 3: cnts = cnts[1]
        else: raise Exception(("Contours tuple must have length 2 or 3, "
                               "otherwise OpenCV changed their cv2.findContours return "
                               "signature yet again. Refer to OpenCV's documentation "
                               "in that case"))
    
        for c in cnts:
            if cv2.contourArea(c) < args.minarea: # minimal changes for detection
                continue
            (x, y, w, h) = cv2.boundingRect(c)
            (x, y, w, h) = (x*2, y*2, w*2, h*2)
            cv2.rectangle(frame, (x, y), (x + w, y + h), (0, 255, 0), 1)
            detected = True

        if detected:
            cv2.putText(frame, 'status: Occupied', (10, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 255), 2)
            if red:
                red.publish('motion', 1)
        else:
            cv2.putText(frame, 'status: Unoccupied', (10, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 0), 2)
        cv2.putText(frame, datetime.datetime.now().strftime("%A %d %B %Y %I:%M:%S%p"), (10, frame.shape[0] - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.35, (0, 0, 255), 1)

        #logging.debug('ProcessImage.processing (time);%0.5f'%(time.time() - start))
        
        if red:
            '''
            start2 = time.time()
            p = pickle.dumps(frame)
            #logging.debug('ProcessImage.pickle.dumps (time);%0.5f'%(time.time() - start2))
            start3 = time.time()
            self.writer.set(('detected', p))
            #logging.debug('ProcessImage.writer.send (time);%0.5f'%(time.time() - start3))

            #start2 = time.time()
            memfile = BytesIO()
            np.save(memfile, frame)
            memfile.seek(0)
            #p = memfile.getvalue()
            #logging.debug('ProcessImage.serialize (time);%0.5f'%(time.time() - start2))
            #start3 = time.time()
            #proc2red.set(('detected', p))
            proc2red.set(('detected', memfile.getvalue()))
            #logging.debug('ProcessImage.writer.send (time);%0.5f'%(time.time() - start3))

            proc2red.set(('detected', pickle.dumps(frame)))
            #logging.warning('asdf')
            '''
            #print(frame.shape, frame.dtype)
            proc2red.set(('detected', frame.tobytes()))
        logging.warning('ProcessImage.overall (time);%0.5f'%(time.time() - start))
